import time

from typing import List
from sktmls import MLSENV, MLSRuntimeENV
from sktmls.models import MLModelClient, AutoMLModel
from sktmls.experiments import Bucket, ExperimentClient


class AutoMLHelper:
    """
    AutoML모델 Helper 클래스 입니다.
    """

    def __init__(
        self, env: MLSENV = None, runtime_env: MLSRuntimeENV = None, username: str = None, password: str = None
    ):
        self.__experiment_client = ExperimentClient(env, runtime_env, username, password)
        self.__model_client = MLModelClient(env, runtime_env, username, password)

    def upgrade_deployment(
        self, prev_model: AutoMLModel, next_model: AutoMLModel, copy_handler: bool = True, stop_prev_model: bool = True
    ) -> List[Bucket]:
        """
        기존 모델 (`prev_model`) 을 신규 모델 (`next_model`) 으로 대체하는 작업을 지원하는 함수입니다. 기존 모델이 연결되어 있는 모든 버킷들에 신규 모델을 적용합니다.

        신규 모델이 동작중이 아닌 경우, 자동으로 동작 중 상태로 업데이트합니다.

        ## Args

        - prev_model: (`sktmls.models.AutoMLModel`) 기존 모델
        - next_model: (`sktmls.models.AutoMLModel`) 신규 모델
        - copy_handler: (optional) (bool) `prev_model`의 핸들러 코드 복제 여부 (`False` 일 경우, 미리 Handler 가 등록되어 있어야 함, Default: `True`)
        - stop_prev_model: (optional) (bool) 업그레이드 완료 후 `prev_model` 의 모델 서버를 중단할지 여부 (Default: `True`)

        ## Example

        ```python
        prev_model = model_client.get_automl_model(name=model_name, version=prev_model_version)
        next_model = model_client.get_automl_model(name=model_name, version=next_model_version)

        automl_helper.upgrade_deployment(prev_model=prev_model, next_model=next_model)
        ```
        """

        def _wait_cache_update():
            time.sleep(60)

        if not self.__model_client.get_automl_model_server_handler(next_model):
            if copy_handler:
                prev_handler = self.__model_client.get_automl_model_server_handler(prev_model)
                if prev_handler:
                    self.__model_client.update_automl_model_server_handler(next_model, prev_handler)
                else:
                    raise Exception(
                        f"'prev_model'에 예측 결과 처리 로직이 등록되어 있지 않습니다 ({prev_model.name} {prev_model.version})."
                    )

            else:
                raise Exception(f"예측 결과 처리 로직이 등록되지 않은 모델입니다 ({next_model.name} {next_model.version}).")

        ams_status = self.__model_client.get_automl_model_server_status(model=next_model)

        start_model = False

        if ams_status in ["STOPPING"]:
            raise Exception(
                f"'{ams_status}' 상태의 모델을 사용할 수 없습니다. 잠시 후 다시 시도해주세요 ({next_model.name} {next_model.version})."
            )
        elif ams_status in ["STOPPED", "FAILED", "UNKNOWN"]:
            self.__model_client.start_automl_model_server(model=next_model)
            time.sleep(60)
            ams_status = self.__model_client.get_automl_model_server_status(model=next_model)
            start_model = True

        wait_tic = 0

        while ams_status == "STARTING" and wait_tic <= 120:
            time.sleep(10)
            wait_tic += 1
            ams_status = self.__model_client.get_automl_model_server_status(model=next_model)

        if ams_status in ["STARTING", "STOPPED"]:
            raise Exception(f"AutoML 모델 동작 Timeout 발생 ({next_model.name} {next_model.version}).")

        elif ams_status in ["FAILED", "UNKNOWN"]:
            raise Exception(f"AutoML 모델 동작 실패: {ams_status} ({next_model.name} {next_model.version}).")

        if start_model:
            _wait_cache_update()

        buckets = self.__experiment_client.migrate_bucket(
            prev_model=prev_model,
            next_prediction_type="automl",
            next_model=next_model,
        )

        if stop_prev_model:
            self.__model_client.stop_automl_model_server(model=prev_model)

        return buckets



import os
from typing import Any, Dict, List

import aiohttp
import requests

from sktmls import MLSENV, MLSRuntimeENV, MLSResponse
from sktmls.config import Config


class MLSRecommendationAPIClient:
    """
    MLS 추천 API를 호출할 수 있는 클라이언트입니다.

    MMS를 제외한 환경에서는 테스트용 호출로 처리됩니다.

    EDD 환경은 지원하지 않습니다.
    """

    def __init__(
        self,
        env: MLSENV = None,
        runtime_env: MLSRuntimeENV = None,
        client_id: str = None,
        apikey: str = None,
    ):
        """
        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.MMS`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - client_id: (optional) (str) 호출 클라이언트 (`tw`, `twd`, `netcrm` 등)
        - apikey: (optional) (str) 클라이언트의 API Key

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $AWS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env

        runtime_env가 `sktmls.MLSRuntimeENV.MMS`인 경우 `client_id`와 `apikey` 파라미터를 생략 가능합니다.

        ## Returns
        `sktmls.apis.MLSRecommendationAPIClient`

        ## Example

        ```python
        recommendation_api_client = MLSRecommendationAPIClient(client_id="tw", apikey="test_apikey",
        env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE)
        ```
        """
        if env:
            assert env in MLSENV.list_items(), "유효하지 않은 MLS 환경입니다."
            self.__env = env
        elif os.environ.get("MLS_ENV"):
            self.__env = MLSENV[os.environ["MLS_ENV"]]
        elif os.environ.get("AWS_ENV"):
            self.__env = MLSENV[os.environ["AWS_ENV"]]
        else:
            self.__env = MLSENV.STG

        if runtime_env:
            assert runtime_env in MLSRuntimeENV.list_items(), "유효하지 않은 런타임 환경입니다."
            self.__runtime_env = runtime_env
        elif os.environ.get("MLS_RUNTINE_ENV"):
            self.__runtime_env = MLSRuntimeENV[os.environ["MLS_RUNTINE_ENV"]]
        else:
            __HOSTNAME = os.environ.get("HOSTNAME", "").lower()
            if __HOSTNAME.startswith("bdp-dmi"):
                self.__runtime_env = MLSRuntimeENV.YE
            elif __HOSTNAME.startswith("vm-skt"):
                self.__runtime_env = MLSRuntimeENV.EDD
            else:
                self.__runtime_env = MLSRuntimeENV.LOCAL

        assert runtime_env != MLSRuntimeENV.EDD, "EDD 환경은 지원하지 않습니다."
        if runtime_env != MLSRuntimeENV.MMS:
            assert client_id and apikey, "`client_id`와 `apikey`가 전달되어야 합니다."
            self.__client_id = client_id
            self.__apikey = apikey
        else:
            self.__client_id = None
            self.__apikey = None

        self.config = Config(self.__runtime_env.value)

    def get_env(self) -> MLSENV:
        return self.__env

    def get_runtime_env(self) -> MLSRuntimeENV:
        return self.__runtime_env

    def request(self, version: str, user_id: str, channel_ids: List[str]) -> Dict[str, Any]:
        """
        MLS 추천 API v1/v2를 호출합니다.

        ## Args

        - version: (str) 호출할 추천 API 버전(`v1`|`v2`)
        - user_id: (str) 해시된 유저 고유 ID (해시된 서비스관리번호)
        - channel_ids: (list(str)) 조회할 채널 ID 리스트

        ## Returns
        dict

        ## Example

        ```python
        rec_dict = recommendation_api_client.request(
            version="v2",
            user_id=hashlib.sha256("1234567890".encode()).hexdigest(),
            channel_ids=["test"]
        )
        ```
        """
        assert version in ["v1", "v2"], "현재 지원하는 버전은 `v1`, `v2`입니다"
        assert channel_ids, "`channel_ids` 리스트가 유효하지 않습니다."

        headers = {"User-Id-Want-Digest": "no"}
        params = {}
        if self.__runtime_env == MLSRuntimeENV.MMS:
            client_id = "mls-mms"
        else:
            client_id = self.__client_id
            headers["apikey"] = self.__apikey
            params["mode"] = "test"

        data = {"user_id": user_id, "channel_ids": channel_ids, "sale_org_id": "admin"}

        response = MLSResponse(
            requests.post(
                url=f"{self.config.MLS_RECOMMENDATION_API_URL[self.__env.value]}/{version}/{client_id}/items",
                headers=headers,
                params=params,
                json=data,
            )
        )
        return response.results

    async def request_async(self, version: str, user_id: str, channel_ids: List[str]) -> Dict[str, Any]:
        """
        MLS 추천 API v1/v2를 비동기 호출합니다.

        ## Args

        - version: (str) 호출할 추천 API 버전(`v1`|`v2`)
        - user_id: (str) 해시된 유저 고유 ID (해시된 서비스관리번호)
        - channel_ids: (list(str)) 조회할 채널 ID 리스트

        ## Returns
        dict

        ## Example

        ```python
        rec_dict = await recommendation_api_client.request_async(
            version="v2",
            user_id=hashlib.sha256("1234567890".encode()).hexdigest(),
            channel_ids=["test"]
        )
        ```
        """
        assert version in ["v1", "v2"], "현재 지원하는 버전은 `v1`, `v2`입니다"
        assert channel_ids, "`channel_ids` 리스트가 유효하지 않습니다."

        headers = {"User-Id-Want-Digest": "no"}
        params = {}
        if self.__runtime_env == MLSRuntimeENV.MMS:
            client_id = "mls-mms"
        else:
            client_id = self.__client_id
            headers["apikey"] = self.__apikey
            params["mode"] = "test"

        data = {"user_id": user_id, "channel_ids": channel_ids, "sale_org_id": "admin"}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f"{self.config.MLS_RECOMMENDATION_API_URL[self.__env.value]}/{version}/{client_id}/items",
                headers=headers,
                json=data,
                params=params,
            ) as response:
                return MLSResponse(status=response.status, body=await response.json()).results

    def predict(self, api_type: str, target_id: str, channel_ids: List[str], keys: List[Any] = []):
        """
        MLS 추천 API v3를 호출합니다.

        ## Args

        - api_type: (str) 식별자 타입 (`user`|`target`)
        - target_id: (str) 추천 타겟의 고유 식별자
        - channel_ids: (list(str)) 조회할 채널 ID 리스트
        - keys: (optional) (list) 추가 전달할 keys 리스트 (기본값: [])

        ## Returns
        dict

        ## Example

        ```python
        rec_dict = recommendation_api_client.predict(
            api_type="user",
            target_id=hashlib.sha256("1234567890".encode()).hexdigest(),
            channel_ids=["test"],
            keys=[],
        )
        ```
        """
        headers = {}
        params = {}
        if self.__runtime_env == MLSRuntimeENV.MMS:
            client_id = "mls-mms"
        else:
            client_id = self.__client_id
            headers["apikey"] = self.__apikey
            params["mode"] = "test"

        data = {"type": api_type, "id": target_id, "channel_ids": channel_ids, "keys": keys}

        response = MLSResponse(
            requests.post(
                url=f"{self.config.MLS_RECOMMENDATION_API_URL[self.__env.value]}/v3/{client_id}/predict",
                headers=headers,
                params=params,
                json=data,
            )
        )
        return response.results

    async def predict_async(self, api_type: str, target_id: str, channel_ids: List[str], keys: List[Any] = []):
        """
        MLS 추천 API v3를 비동기 호출합니다.

        ## Args

        - api_type: (str) 식별자 타입 (`user`|`target`)
        - target_id: (str) 추천 타겟의 고유 식별자
        - channel_ids: (list(str)) 조회할 채널 ID 리스트
        - keys: (optional) (list) 추가 전달할 keys 리스트 (기본값: [])

        ## Returns
        dict

        ## Example

        ```python
        rec_dict = await recommendation_api_client.predict_async(
            api_type="user",
            target_id=hashlib.sha256("1234567890".encode()).hexdigest(),
            channel_ids=["test"],
            keys=[],
        )
        ```
        """
        headers = {}
        params = {}
        if self.__runtime_env == MLSRuntimeENV.MMS:
            client_id = "mls-mms"
        else:
            client_id = self.__client_id
            headers["apikey"] = self.__apikey
            params["mode"] = "test"

        data = {"type": api_type, "id": target_id, "channel_ids": channel_ids, "keys": keys}

        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f"{self.config.MLS_RECOMMENDATION_API_URL[self.__env.value]}/v3/{client_id}/predict",
                headers=headers,
                json=data,
                params=params,
            ) as response:
                return MLSResponse(status=response.status, body=await response.json()).results



from types import FunctionType
import marshal
import ipyparallel.serialize.codeutil  # noqa: F401

import inspect

check_commend = [
    "rm",
    "rm -rf",
    "mv",
    "-f",
    "-i",
    "-u",
    "-v",
    "-b",
    "echo",
    "cd",
    "cp",
    "scp",
    "chmod",
    "ls",
    "import os",
    "import subprocess",
    "import sys",
    "import asyncio",
    "!",
    "&",
    "--y",
    "sudo",
    "passwd",
    "getpass",
]


class LogicConverter:
    """
    상품 선택 로직 함수를 변환하거나 변환된 함수를 다시 재변환을 지원하는 클래스입니다.

    ## Example

    ```python
    def Handler(data = None):
        if data:
            if data['prod_info'][0]['bas_fee_amt'] < 50000:
                return [{"id": "prod_id"
                         , "name": "prod_nm"
                         , "type": "vas"
                         , "props": {"class": "sub"
                                     , "priority" : 'Y'
                                     , "fee_prod_id" : data['fee_prod_id']
                                     , "score" : data['y'][1]}}]
            else:
                return [{"id": "prod_id"
                         , "name": "prod_nm"
                         , "type": "vas"
                         , "props": {"class": "sub"
                                     , "priority" : 'N'
                                     , "fee_prod_id" : data['fee_prod_id']
                                     , "score" : data['y'][1]}}]

        else:
            return [{"id": "prod_id"
                     , "name": "prod_nm"
                     , "type": "vas"
                     , "props": {"class": None
                                 , "priority" : None
                                 , "fee_prod_id" : None
                                 , "score" : None}}]

    func_convert_value = ConvertVariables(user_defined_func = Handler).FunctionToConvert() #변환
    reconvert_value = ConvertVariables(user_defined_convert_logic = func_convert_value).ConvertToFunction() #재변환
    """

    def __init__(
        self,
        user_defined_func=None,
        user_converted_logic=None,
    ):
        if user_defined_func:
            assert hasattr(user_defined_func, "__call__") is True, "`user_defined_func`은 함수이어야 합니다."
            for i in check_commend:
                for j in inspect.getsource(user_defined_func).split("\n"):
                    assert i is not j, f"금지된 명령어 {i}가 {j}로 입력되었습니다."
        if user_converted_logic:
            assert isinstance(user_converted_logic, bytes), "`user_converted_logic`는 LogicConverter를 거쳐 반환된 타입이어야 합니다."

        self.user_defined_func = user_defined_func
        self.user_converted_logic = user_converted_logic

    def dummy_handler(self):
        return {"id": "id", "name": "name", "type": "type", "props": {}}, ""

    def FunctionToConvert(self):
        func_to_bytes = marshal.dumps(self.user_defined_func.__code__)
        return func_to_bytes

    def ConvertToFunction(self, name=None):
        if self.user_converted_logic:
            assert (
                hasattr(FunctionType(marshal.loads(self.user_converted_logic), globals={}, name=name), "__call__")
                is True
            ), "`user_converted_logic`의 변환 값은 함수이어야 합니다."
            bytes_to_function = FunctionType(marshal.loads(self.user_converted_logic), globals={}, name=name)
            return bytes_to_function
        else:
            return self.dummy_handler()



from sktmls.apis import MLSProfileAPIClient
from typing import Dict, List, Any
from sktmls import MLSENV, MLSRuntimeENV
from sktmls.utils import LogicConverter
from sktmls.models import MLSModelError
import hvac
import time
import numpy as np


class EurekaModelTest(MLSProfileAPIClient):
    def __init__(
        self,
        profile_id: str,
        model,
        client_id: str = "netcrm",
        apikey: str = "DGHIIVU4PS4FI9ECJJ7QWHKN9OS8OHVGR1S961YC",
        env: str = "stg",
        runtime_env: str = "ye",
        additional_keys: str = None,
    ):
        """
        AutoGluon을 통해 학습한 모델의 테스트를 지원합니다.
        ## Example
        # 기학습 모델 준비
        model_test = EurekaModelTest(
            profile_id='user_profile_smp_result',
            model = kyu_test_model,
            client_id = "netcrm",
            apikey = "DGHIIVU4PS4FI9ECJJ7QWHKN9OS8OHVGR1S961YC",
            env = "stg",
            runtime_env = "ye",
        )
        # prediction & 후처리 테스트 Example
        result = model_test.test_logic(user_id = user_id)
        print(result)
        [{'id': 'prod_id', 'name': 'prod_nm', 'type': 'vas', 'props': {'class': 'sub', 'fee_prod_id': 'NA00006402', 'score': 0.4414701759815216}}]

        #전처리 결과
        user_profile_dict, preprocessed_features = model_test._preprocess_logic_test(user_id = user_id)
        #예측 결과
        predictions = model_test._ml_prediction_test(preprocessed_features = preprocessed_features)
        #후처리 결과
        result = model_test._postprocess_logic_test(user_profile_dict = user_profile_dict
                                                   , user_id = user_id
                                                   , additional_keys = None
                                                   , predictions = predictions)
        print(result)

        user_profile_dict, preprocessed_features, predictions = model_test._ml_predict(user_id=user_id)
        print(predictions)
        [0.7659122347831726, 0.2340877503156662]
        # 후처리 테스트
        results= model_test._postprocess_logic_test(user_profile_dict = user_profile_dict, user_id = user_id, predictions = predictions)
        print(results)
        [{'id': 'prod_id', 'name': 'prod_nm', 'type': 'vas', 'props': {'class': 'sub', 'fee_prod_id': 'NA00006402', 'score': 0.4414701759815216}}]
        """
        assert type(env) is str, "테스트 환경 값은 String 이어야 합니다.(stg, dev, prd)"
        assert type(runtime_env) is str, "runtime_env 값은 String 이어야 합니다.(ye, mms etc)"
        assert type(profile_id) is str, "profile_id 값은 String 이어야 합니다.(user_profile_smp_result etc)"
        assert type(client_id) is str, "client_id 값은 String 이어야 합니다.(netcrm etc)"
        assert type(apikey) is str, "apikey 값은 String 이어야 합니다. "
        self.env = env
        self.profile_id = profile_id
        self.model = model
        self.pf_client = MLSProfileAPIClient(
            env=MLSENV.PRD if env == "prd" else MLSENV.STG,
            runtime_env=MLSRuntimeENV.MMS if runtime_env == "mms" else MLSRuntimeENV.YE,
            client_id=client_id,
            apikey=apikey,
        )
        self.additional_keys = additional_keys

    def _preprocess_logic_test(
        self,
        user_id: str,
    ) -> List[Any]:
        assert type(user_id) is str, "테스트하고자 하는 유저 ID 값은 str 이어야 합니다."

        user_profile_dict = self.pf_client.get_user_profile(
            profile_id=self.profile_id,
            user_id=user_id,
            keys=self.model.features,
        )

        # features 순서에 맞게 리스트로 정리
        feature_values = [user_profile_dict[feature_name] for feature_name in self.model.features]
        # Preprocessing
        preprocessed_features = self.model._preprocess(
            x=feature_values,
            additional_keys=self.additional_keys,
            pf_client=self.pf_client,
            dynamodb_client=None,
            graph_client=None,
        )
        return user_profile_dict, preprocessed_features

    def _ml_prediction_test(
        self,
        preprocessed_features,
    ) -> List[Any]:
        """
        AutoGluon을 통해 학습한 모델의 전처리 및 예측 결과를 반환합니다.
        ## Example
        # 학습 및 테스트 데이터 준비
        model_test = MLSModelTest(
            profile_id=profile_id,
            model = model_name,
            client_id = client_id,
            apikey = apikey,
            env = "stg",
            runtime_env = "ye",
        )
        # prediction Example
        user_profile_dict, preprocessed_features, predictions = model_test._ml_predict(user_id=user_id)
        print(predictions)
        [0.7659122347831726, 0.2340877503156662]
        """
        # ML Prediction
        predictions = self.model._ml_predict(preprocessed_x=preprocessed_features)

        return predictions

    def _postprocess_logic_test(
        self, user_profile_dict: Dict[str, Any], user_id: str, additional_keys: str, predictions: Dict[str, Any] = None
    ):
        """
        AutoGluon을 통해 학습한 모델의 후처리 결과를 반환합니다.
        ## 참고 사항
        - `postprocess_logic_func` 함수를 통해 동작하므로 postprocess_logic_func 는 함수여야 합니다.
        ## 후처리 Example
        results= model_test._postprocess_logic_test(user_profile_dict = user_profile_dict, user_id = user_id, predictions = predictions)
        print(results)
        [{'id': 'prod_id', 'name': 'prod_nm', 'type': 'vas', 'props': {'class': 'sub', 'fee_prod_id': 'NA00006402', 'score': 0.4414701759815216}}]
        """
        assert isinstance(user_profile_dict, dict), "`user_profile_dict`는 dict 타입이어야 합니다."
        assert type(user_id) is str, "테스트하고자 하는 유저 ID 값은 str 이어야 합니다."
        assert type(predictions) is list, "테스트하고자 하는 predictions 값은 list 이어야 합니다."
        data = user_profile_dict
        data["additional_keys"] = self.additional_keys
        data["y"] = predictions
        data["user_id"] = user_id

        postprocessing_user_profile = self.model.postprocessing_user_profile
        if postprocessing_user_profile:
            for user_profile in postprocessing_user_profile:
                user_values = self.pf_client.get_user_profile(
                    profile_id=user_profile, user_id=data["user_id"], keys=postprocessing_user_profile[user_profile]
                )
                assert (
                    None not in user_values.values()
                ), "후처리 시 사용되는 `user profile` 변수의 값은 사용되는 `user profile`에 있어야 합니다."
                data.update(user_values)

        postprocessing_item_profile_preprocessed = self.model._get_var(
            data=data, postprocessing_item_profile=self.model.postprocessing_item_profile
        )

        if postprocessing_item_profile_preprocessed:
            for item_profile in postprocessing_item_profile_preprocessed:
                for item_id, keys in postprocessing_item_profile_preprocessed[item_profile].items():
                    item_values = self.pf_client.get_item_profile(profile_id=item_profile, item_id=item_id, keys=keys)
                    data.update(item_values)
                    assert (
                        None not in user_values.values()
                    ), "후처리 시 사용되는 `item profile` 변수의 값은 사용되는 `item profile`에 있어야 합니다."
        try:
            if self.model.postprocess_logic is not None:
                assert isinstance(
                    self.model.postprocess_logic, bytes
                ), "`postprocess_logic`은 Logic Converter를 통해 변환된 타입이어야 합니다."
                Handler = LogicConverter(user_converted_logic=self.model.postprocess_logic).ConvertToFunction(
                    name=f"{self.model.model_name}_{self.model.model_version}_Handler"
                )
                assert hasattr(Handler, "__call__") is True, "후처리 로직은 함수이어야 합니다."
                return Handler(data)
        except Exception as e:
            raise MLSModelError(f"EurekaModel: 후처리에 실패했습니다. {e}")

    def test_logic(
        self,
        user_id: str,
    ) -> Any:
        """
        AutoGluon을 통해 학습한 모델의 테스트를 지원합니다.
        # prediction & 후처리 테스트 Example
        result = model_test.test_logic(user_id = user_id)
        print(result)
        [{'id': 'prod_id', 'name': 'prod_nm', 'type': 'vas', 'props': {'class': 'sub', 'fee_prod_id': 'NA00006402', 'score': 0.4414701759815216}}]
        """

        user_profile_dict, preprocessed_features = self._preprocess_logic_test(user_id=user_id)
        predictions = self._ml_prediction_test(preprocessed_features=preprocessed_features)
        return self.model._postprocess(
            x=preprocessed_features + [user_id],
            additional_keys=self.additional_keys,
            y=predictions,
            pf_client=self.pf_client,
            graph_client=None,
            dynamodb_client=None,
        )


def get_secrets(path, parse_data=True):
    vault_client = hvac.Client()
    data = vault_client.secrets.kv.v2.read_secret_version(path=path)
    if parse_data:
        data = data["data"]["data"]
    return data


def generate_configs(env, user):
    import pkg_resources
    from packaging import version
    from sktmls import MLSENV

    mlsenv = MLSENV.STG
    if env == "prd":
        mlsenv = MLSENV.PRD
    secrets = get_secrets(path="mls")
    config = dict(
        env=mlsenv,
        username=secrets.get(f"{user}_id"),
        password=secrets.get(f"{user}_pass"),
    )

    # sktmls version upgrade 후 수정
    mls_v = pkg_resources.get_distribution("sktmls").version
    if version.parse(mls_v) > version.parse("2020.8.29"):
        from sktmls import MLSRuntimeENV

        return [{**config, "runtime_env": r} for r in MLSRuntimeENV.list_items() if r != MLSRuntimeENV.EDD]

    return [config]


def get_mls_config(env, user):
    import concurrent.futures

    configs = generate_configs(env=env, user=user)
    e = concurrent.futures.ThreadPoolExecutor(max_workers=len(configs) + 1)
    fs = [e.submit(check_client_config, conf) for conf in configs]
    for f in concurrent.futures.as_completed(fs):
        if f.exception() is None:
            config = f.result()
            break
    e.shutdown(wait=False)
    return config


def check_client_config(config):
    from sktmls.meta_tables.meta_table import MetaTableClient

    client = MetaTableClient(**config)
    client.list_meta_tables()
    return config


def get_mls_ml_model_client(env="stg", user="aa2"):
    from sktmls.models import MLModelClient

    config = get_mls_config(env, user)
    return MLModelClient(**config)


def get_mls_experiment_client(env="stg", user="aa2"):
    from sktmls.experiments import ExperimentClient

    config = get_mls_config(env, user)
    return ExperimentClient(**config)


def get_mls_profile_api_client(env="stg", user="aa2"):
    from sktmls.apis.profile_api import MLSProfileAPIClient

    config = dict({conf for conf in get_mls_config(env, user).items() if conf[0] in ["env", "runtime_env"]})
    return MLSProfileAPIClient(**config, **get_secrets("mls").get("reco_api"))


def get_mls_recommendation_api_client(env="stg", user="aa2"):
    from sktmls.apis.recommendation_api import MLSRecommendationAPIClient

    config = dict({conf for conf in get_mls_config(env, user).items() if conf[0] in ["env", "runtime_env"]})
    return MLSRecommendationAPIClient(**config, **get_secrets("mls").get("reco_api"))


class EurekaExpModelTest:
    def __init__(
        self,
        env: str = "stg",
        user: str = "aa2",
        model_name: str = "kyu_auto_test_model",
        model_version: str = "v1",
        test_experiment: str = "kyu_test_model",
        test_channel: str = "kyu_test_model",
        bucket_name: str = "A",
    ):
        """
        AutoGluon을 통해 학습한 모델의 실제 MLS 연결 테스트를 지원합니다.
        ## Example
        exp_test = EurekaExpModelTest(env= "stg",
            user= "aa2",
            model_name = "kyu_auto_test_model",
            model_version = "v1",
            test_experiment = "kyu_test_model",
            test_channel = "kyu_test_model",
            bucket_name = "A"
        )

        # 실제 실험 연결 및 테스트
        test_info, success_svc, error_svc, running_time = exp_test.model_test_real_experiments(user_ids= user_ids, prediction_type='online')
        print(test_info)
        {'NUMBER_OF_NO_USER_PROFILE': 34, 'NUMBER_OF_SUCCESS': 65, 'NUMBER_OF_EMPTY_RESPONSE': 0, 'NUMBER_OF_ERROR': 0, 'RUNNING_TIME_AVG': 0.2112327428964468}

        #실제 실험 연결
        profile_api_client, recommendation_api_client = exp_test.test_model_connect_experiment(prediction_type = "online", profile_id = "user_default")
        """
        assert type(env) is str, "테스트 환경 값은 String 이어야 합니다.(stg, prd)"
        assert type(user) is str, "user 값은 String 이어야 합니다.(aa2 etc)"
        assert type(model_name) is str, "model_name 값은 String 이어야 합니다.(kyu_auto_test_model etc)"
        assert type(model_version) is str, "model_version 값은 String 이어야 합니다.(v1 etc)"
        assert type(test_experiment) is str, "test_experiment 값은 String 이어야 합니다.(kyu_test_model etc)"
        assert type(test_channel) is str, "test_channel 값은 String 이어야 합니다.(kyu_test_model etc)"
        assert type(bucket_name) is str, "bucket_name 값은 String 이어야 합니다.(A etc)"
        self.env = env
        self.user = user
        self.model_name = model_name
        self.model_version = model_version
        self.test_experiment = test_experiment
        self.test_channel = test_channel
        self.bucket_name = bucket_name

    def test_model_connect_experiment(self, prediction_type="online"):
        model_client = get_mls_ml_model_client(env=self.env, user=self.user)
        model = model_client.get_manual_model(name=self.model_name, version=self.model_version)

        experiment_client = get_mls_experiment_client(env=self.env, user=self.user)
        my_experiment = experiment_client.get_experiment(name=self.test_experiment)

        my_bucket = experiment_client.get_bucket(experiment=my_experiment, name=self.bucket_name)

        experiment_client.update_bucket(bucket=my_bucket, prediction_type=prediction_type, model=model)

        recommendation_api_client = get_mls_recommendation_api_client(env=self.env, user=self.user)
        profile_api_client = get_mls_profile_api_client(env=self.env, user=self.user)
        return profile_api_client, recommendation_api_client

    def model_test_real_experiments(
        self, user_ids: List[str], prediction_type: str = "online", profile_id: str = "user_default"
    ) -> List[Any]:
        assert isinstance(user_ids, list), "`user_id`은 list 타입이어야 합니다."
        assert type(prediction_type) is str, "`prediction_type`은 str 타입이어야 합니다."
        assert type(profile_id) is str, "`profile_id`은 str 타입이어야 합니다."

        profile_api_client, recommendation_api_client = self.test_model_connect_experiment(
            prediction_type=prediction_type
        )

        no_user_profile_cnt = 0
        success_cnt = 0
        empty_reponse_cnt = 0
        error_cnt = 0

        error_svc = []
        success_svc = []

        running_time = []

        for i in range(0, len(user_ids)):
            try:
                profile_api_client.get_user_profile(profile_id=profile_id, user_id=user_ids[i], keys=["user_id"])[
                    "user_id"
                ]
                try:
                    start_time = time.time()
                    response = recommendation_api_client.predict(
                        api_type="user",
                        target_id=user_ids[i],
                        channel_ids=[self.test_channel],
                    )[self.test_channel]
                    end_time = time.time()
                    running_time.append(end_time - start_time)
                    if len(response) < 1:
                        empty_reponse_cnt = empty_reponse_cnt + 1
                    else:
                        success_svc.append(user_ids[i])
                        success_cnt = success_cnt + 1
                except Exception as e:
                    error_svc.append(user_ids[i])
                    print(f"error svc: {user_ids[i]}")
                    print(e)
                    error_cnt = error_cnt + 1
            except Exception:
                no_user_profile_cnt = no_user_profile_cnt + 1

        test_info = {
            "NUMBER_OF_NO_USER_PROFILE": no_user_profile_cnt,
            "NUMBER_OF_SUCCESS": success_cnt,
            "NUMBER_OF_EMPTY_RESPONSE": empty_reponse_cnt,
            "NUMBER_OF_ERROR": error_cnt,
            "RUNNING_TIME_AVG": np.mean(running_time),
        }
        return test_info, success_svc, error_svc, running_time



import numpy as np

from sktmls import MLSRuntimeENV
from sktmls.apis import MLSProfileAPIClient, MLSGraphAPIClient
from sktmls.dynamodb import DynamoDBClient
from sktmls.models.contrib import EurekaModel


class TestEurekaModel:
    def test_000_generic_context_success(self, mocker):
        mocker.patch(
            "sktmls.models.contrib.eureka_model.EurekaModel._ml_predict",
            return_value=np.array([0, 1]),
        )
        mocker.patch("sktmls.apis.MLSProfileAPIClient.get_item_profile", return_value={"test_dimension": 300})
        mocker.patch(
            "sktmls.apis.MLSGraphAPIClient.list_vertices",
            return_value=[{"id": "test001", "label": "testlabel", "properties": {"hello": "world"}}],
        )
        mocker.patch("sktmls.dynamodb.DynamoDBClient.get_item", return_value={"id": "ITEM0001", "age": 3})

        def Handler(data=None):
            if data:
                return {
                    "id": "prod_id",
                    "name": "prod_nm",
                    "type": "vas",
                    "props": {"class": "sub", "score": data["y"][1]},
                }

            else:
                return []

        automl = EurekaModel(
            model_name="test_model",
            model_version="test_version",
            features=["feature1", "feature2", "feature3", "emb_feature1", "context_feature1", "context_feature2"],
            preprocess_logic=None,
            postprocess_logic=Handler,
            predict_fn="predict",
            use_logic_processer="N",
        )

        assert automl._preprocess(
            [9.0, 10.0, 2.0, 4.0, None, "Y"],
            [1],
            MLSProfileAPIClient(client_id="a", apikey="a"),
            MLSGraphAPIClient(),
            DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == [9.0, 10.0, 2.0, 4.0, None, "Y"]

        assert automl._postprocess(
            x=[9.0, 10.0, 2.0, 4.0, 8.0, "Y"],
            additional_keys=[1],
            y=[0, 3],
            pf_client=MLSProfileAPIClient(client_id="a", apikey="a"),
            graph_client=MLSGraphAPIClient(),
            dynamodb_client=DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == [{"id": "prod_id", "name": "prod_nm", "type": "vas", "props": {"class": "sub", "score": 3}}]

        assert automl.predict([9.0, 10.0, 2.0, 4.0, None, "Y"],) == {
            "items": [
                {
                    "id": "prod_id",
                    "name": "prod_nm",
                    "type": "vas",
                    "props": {"class": "sub", "score": 1},
                }
            ]
        }

    def test_001_generic_context_success(self, mocker):
        mocker.patch(
            "sktmls.models.contrib.eureka_model.EurekaModel._ml_predict",
            return_value=np.array([0, 1]),
        )
        mocker.patch("sktmls.apis.MLSProfileAPIClient.get_item_profile", return_value={"test_dimension": 300})
        mocker.patch(
            "sktmls.apis.MLSGraphAPIClient.list_vertices",
            return_value=[{"id": "test001", "label": "testlabel", "properties": {"hello": "world"}}],
        )
        mocker.patch("sktmls.dynamodb.DynamoDBClient.get_item", return_value={"id": "ITEM0001", "age": 3})

        def Handler(data=None):
            if data:
                if data["context_feature2"] < 3:
                    return {
                        "id": "prod_id",
                        "name": "prod_nm",
                        "type": "vas",
                        "props": {"class": "sub", "score": data["y"][1]},
                    }
                else:
                    return {
                        "id": "prod_id",
                        "name": "prod_nm",
                        "type": "vas",
                        "props": {"class": "sub", "score": data["y"][1]},
                    }

            else:
                return {
                    "id": "prod_id",
                    "name": "prod_nm",
                    "type": "vas",
                    "props": {"class": "sub", "score": "None"},
                }

        automl = EurekaModel(
            model_name="test_model",
            model_version="test_version",
            features=["feature1", "feature2", "feature3", "emb_feature1", "context_feature1", "context_feature2"],
            preprocess_logic=None,
            postprocess_logic=Handler,
            predict_fn="predict",
            use_logic_processer="N",
        )

        assert automl._preprocess(
            [9.0, 10.0, 10.0, None, "Y", 0],
            [1],
            MLSProfileAPIClient(client_id="a", apikey="a"),
            MLSGraphAPIClient(),
            DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == [9.0, 10.0, 10.0, None, "Y", 0]

        assert automl._postprocess(
            x=[9.0, 10.0, 10.0, 2.0, "Y", 10.0],
            additional_keys=[1],
            y=[0, 3],
            pf_client=MLSProfileAPIClient(client_id="a", apikey="a"),
            graph_client=MLSGraphAPIClient(),
            dynamodb_client=DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == [
            {
                "id": "prod_id",
                "name": "prod_nm",
                "type": "vas",
                "props": {"class": "sub", "score": 3},
            }
        ]

        assert automl.predict([9.0, 10.0, 10.0, 2.0, "Y", 1.0],) == {
            "items": [
                {
                    "id": "prod_id",
                    "name": "prod_nm",
                    "type": "vas",
                    "props": {"class": "sub", "score": 1},
                }
            ]
        }



from typing import Any, Dict, List, Union

import numpy as np
from pandas import DataFrame
from pytz import timezone
import torch

from sktmls import MLSRuntimeENV
from sktmls.apis import MLSProfileAPIClient, MLSGraphAPIClient
from sktmls.dynamodb import DynamoDBClient
from sktmls.models import MLSGenericModel, MLSModelError
from sktmls.utils import LogicProcessor

TZ = timezone("Asia/Seoul")

logic_processor = LogicProcessor()


class ConditionalGenericLogicModel(MLSGenericModel):
    """
    MLS 모델 레지스트리에 등록되는 다수 모델 기반의 클래스입니다.

    모델 선택 로직으로 적용할 모델을 선택한 뒤 전처리 로직과 후처리 로직을 기반으로 프로세스합니다.
    """

    def __init__(
        self,
        model_name: str,
        model_version: str,
        features: List[str],
        models: List[Any],
        preprocess_logics: List[Dict[str, List[Any]]],
        postprocess_logics: List[Dict[str, List[Any]]],
        predict_fns: List[str],
        model_selection_logic: Dict[str, List[Any]] = {"abs": 0},
        data: Dict[str, Any] = {},
    ):
        """
        `sktmls.models.contrib.GenericLogicModel`과 동작 원리는 동일하나, `models`에 여러 개의 ML 라이브러리로 학습된 모델을 저장해 두고 `model_selection_logic`을 통해 하나를 선택하여 predict를 수행합니다.

        ## Args

        - model_name: (str) 모델 이름
        - model_version: (str) 모델 버전
        - features: (list(str)) 피쳐 이름 리스트
        - models: ML 라이브러리로 학습한 모델 객체 리스트
        - preprocess_logics: (list(dict)) 각 모델에 대해 전달된 피쳐 `x`로부터 ML모델의 predict 함수에 input으로 들어갈 `preprocessed_x`를 만드는 전처리 로직 리스트
        - postprocess_logics: (list(dict)) 각 ML모델의 predict 함수 결과로 얻어진 `y`로부터 리턴 body(items 리스트)를 만드는 후처리 로직 리스트
        - predict_fns: (list(str)) ML모델의 추론 함수 이름 (`predict`|`predict_proba`|`none`)
            - `none` 전달 시 ML모델 추론을 사용하지 않습니다 (룰 모델).
            - `model_lib`이 `pytorch` 인 경우 이 파라미터는 무시됩니다.
        - model_selection_logic: (dict) `models`의 모델 중 하나를 선택하는 로직 (기본값: {"abs": 0} (models[0] 모델 사용))
            - 정수 반환 시 models의 인덱스로 사용합니다.
            - list(dict) 반환 시 즉시 그 값을 리턴합니다. 이 때 유효한 response 형식이어야 합니다.
        - data: (optional) (dict) 각 모델의 `preprocess_logic`과 `postprocess_logic`에서 피쳐와 함께 "var" 참조할 추가 데이터 (기본값: {})
            - 피쳐 이름과 같은 키 존재 시 피쳐 값을 덮어쓰게 됩니다. 주의하세요!

        `models`, `preprocess_logics`, `postprocess_logics`, `predict_fns`는 `sktmls.models.contrib.GenericLogicModel`의 `model`, `preprocess_logic`, `postprocess_logic`, `predict_fn`을 리스트로 여러 개 가지고 있는 형태입니다.

        자세한 형식은 `sktmls.models.contrib.GenericLogicModel` 문서를 참조하세요.

        ## Example

        my_model_v1 = ConditionalGenericLogicModel(
            model_name="my_model",
            model_version="v1",
            features=["feature1", "feature2", "feature3", "embedding_vector", "context_feature1", "context_feature2"],
            models=[model0, model1],
            preprocess_logics=[preprocess_logic0, preprocess_logic1],
            postprocess_logics=[postprocess_logic0, postprocess_logic1],
            predict_fns=["predict", "predict_proba"],
            model_selection_logic={
                "if": [
                    {">": [{"var": "feature1"}, 3.0]},
                    0,
                    1
                ]
            },
            data={}
        )
        """
        assert isinstance(features, list), "`features`은 list 타입이어야 합니다."

        assert isinstance(model_selection_logic, dict), "`model_selection_logic`은 dict 타입이어야 합니다."
        for key in model_selection_logic.keys():
            assert key in ["var", "missing", "missing_some", "pf", "with", "for"] or key in logic_processor.operations

        assert (
            len(models) == len(preprocess_logics) == len(postprocess_logics) == len(predict_fns)
        ), "`preprocess_logics`, `postprocess_logics`, `predict_fns`의 길이는 같아야 합니다."

        assert isinstance(preprocess_logics, list), "`preprocess_logics`는 list 타입이어야 합니다."
        for preprocess_logic in preprocess_logics:
            assert isinstance(preprocess_logic, dict), "`preprocess_logic`은 dict 타입이어야 합니다."
            for key in preprocess_logic.keys():
                assert (
                    key in ["var", "missing", "missing_some", "pf", "with", "for"] or key in logic_processor.operations
                )

        assert isinstance(postprocess_logics, list), "`postprocess_logics`는 list 타입이어야 합니다."
        for postprocess_logic in postprocess_logics:
            assert isinstance(postprocess_logic, dict), "`postprocess_logic`은 dict 타입이어야 합니다."
            for key in postprocess_logic.keys():
                assert (
                    key in ["var", "missing", "missing_some", "pf", "with", "for"] or key in logic_processor.operations
                )

        assert isinstance(predict_fns, list), "`predict_fn`은 list 타입이어야 합니다."
        for predict_fn in predict_fns:
            assert predict_fn in [
                "predict",
                "predict_proba",
                "none",
            ], "`predict_fn`은 predict, predict_proba, none 중 하나의 값이어야 합니다."

        assert isinstance(data, dict), "`data`는 dict 타입이어야 합니다."

        super().__init__(models, model_name, model_version, features)

        self.preprocess_logics = preprocess_logics
        self.postprocess_logics = postprocess_logics
        self.predict_fns = predict_fns
        self.model_selection_logic = model_selection_logic
        self.data = data

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        if len(self.features) != len(x):
            raise MLSModelError("ConditionalGenericLogicModel: `x`의 길이가 `features`의 길이와 다릅니다.")

        pf_client = kwargs.get("pf_client") or MLSProfileAPIClient(runtime_env=MLSRuntimeENV.MMS)
        graph_client = kwargs.get("graph_client") or MLSGraphAPIClient(runtime_env=MLSRuntimeENV.MMS)
        dynamodb_client = kwargs.get("dynamodb_client") or DynamoDBClient(runtime_env=MLSRuntimeENV.MMS)

        result = self._select_model(x, kwargs.get("keys", []), pf_client, graph_client, dynamodb_client)
        if isinstance(result, int):
            if result < 0 or result >= len(self.models):
                raise MLSModelError(f"ConditionalGenericLogicModel: model_index({result})가 유효하지 않습니다.")
            model_index = result
        elif isinstance(result, list):
            return {"items": result}
        else:
            raise MLSModelError("ConditionalGenericLogicModel: `model_selection_logic`이 유효하지 않습니다.")

        preprocessed_x = self._preprocess(
            model_index, x, kwargs.get("keys", []), pf_client, graph_client, dynamodb_client
        )
        y = self._ml_predict(model_index, preprocessed_x)
        items = (
            self._postprocess(model_index, x, kwargs.get("keys", []), y, pf_client, graph_client, dynamodb_client) or []
        )

        return {"items": items}

    def _select_model(
        self,
        x: List[Any],
        additional_keys: List[Any],
        pf_client: MLSProfileAPIClient,
        graph_client: MLSGraphAPIClient,
        dynamodb_client: DynamoDBClient,
    ) -> Union[int, List[Dict[str, Any]]]:
        data = {name: x[i] for i, name in enumerate(self.features) if x[i] not in [None, []]}
        data["additional_keys"] = additional_keys
        data.update(self.data)

        try:
            return logic_processor.apply(
                self.model_selection_logic,
                data=data,
                pf_client=pf_client,
                graph_client=graph_client,
                dynamodb_client=dynamodb_client,
            )
        except Exception as e:
            raise MLSModelError(f"ConditionalGenericLogicModel: 모델 선택에 실패했습니다. {e}")

    def _preprocess(
        self,
        model_index: int,
        x: List[Any],
        additional_keys: List[Any],
        pf_client: MLSProfileAPIClient,
        graph_client: MLSGraphAPIClient,
        dynamodb_client: DynamoDBClient,
    ) -> List[Any]:
        data = {name: x[i] for i, name in enumerate(self.features) if x[i] not in [None, []]}
        data["additional_keys"] = additional_keys
        data["model_index"] = model_index
        data.update(self.data)

        try:
            return logic_processor.apply(
                self.preprocess_logics[model_index],
                data=data,
                pf_client=pf_client,
                graph_client=graph_client,
                dynamodb_client=dynamodb_client,
            )
        except Exception as e:
            raise MLSModelError(f"ConditionalGenericLogicModel: 전처리에 실패했습니다. {e}")

    def _ml_predict(self, model_index: int, preprocessed_x: List[Any]) -> Union[float, List[float], str, None]:
        try:
            if self.predict_fns[model_index] == "none" and self.model_lib != "pytorch":
                return None

            if not isinstance(preprocessed_x[0], list):
                preprocessed_x = [preprocessed_x]

            if self.model_lib == "autogluon":
                input_data = DataFrame(
                    preprocessed_x, columns=[f for f in self.features if f not in self.non_training_features]
                )
            elif self.model_lib == "pytorch":
                input_data = torch.tensor(preprocessed_x, dtype=torch.float)
            else:
                input_data = np.array(preprocessed_x)

            if self.model_lib == "pytorch":
                y = self.models[model_index](input_data).detach().numpy()
            elif self.predict_fns[model_index] == "predict":
                y = self.models[model_index].predict(input_data)
            else:
                y = self.models[model_index].predict_proba(input_data)

            if len(y) == 1:
                y = y[0]

            try:
                return y.tolist()
            except AttributeError:
                return y

        except Exception as e:
            raise MLSModelError(f"ConditionalGenericLogicModel: ML Prediction에 실패했습니다. {e}")

    def _postprocess(
        self,
        model_index: int,
        x: List[Any],
        additional_keys: List[Any],
        y: Union[float, List[float], None],
        pf_client: MLSProfileAPIClient,
        graph_client: MLSGraphAPIClient,
        dynamodb_client: DynamoDBClient,
    ) -> List[Dict[str, Any]]:
        data = {name: x[i] for i, name in enumerate(self.features) if x[i] not in [None, []]}
        data["additional_keys"] = additional_keys
        data["model_index"] = model_index
        data["y"] = y
        data.update(self.data)

        try:
            return logic_processor.apply(
                self.postprocess_logics[model_index],
                data=data,
                pf_client=pf_client,
                graph_client=graph_client,
                dynamodb_client=dynamodb_client,
            )
        except Exception as e:
            raise MLSModelError(f"ConditionalGenericLogicModel: 후처리에 실패했습니다. {e}")



import re
import sys

from dateutil import parser
from typing import Any, Dict, Union

from sktmls import MLSClient, MLSENV, MLSRuntimeENV
from sktmls.datasets import DatasetClient, ProblemType
from sktmls.models import AutoMLModel


class AutoMLPredictionError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ScoreExportOption:
    def __init__(self, sort: bool, target_limit: int = None):
        """
        스코어 출력 옵션 값입니다(스코어 기반 고객 타겟팅 문제만 지원).

        ## Args

        - sort: (bool) 결과 정렬 여부
        - target_limit: (int) 총 출력 개수 (sort = True 인 경우 필수)

        ## Returns
        `sktmls.models.automl.ScoreExportOption`

        ## Example

        ```python
        score_export_option = ScoreExportOption(sort=True, target_limit=1000000)
        ```
        """
        assert type(sort) == bool, "sort 은(는) bool 타입이어야 합니다."
        assert type(target_limit) == int if target_limit is not None else True, "target_limit 은(는) int 타입이어야 합니다."
        assert target_limit is not None if sort else True, "sort = True 인 경우 target_limit 은 반드시 전달되어야 합니다."
        assert target_limit <= 10000000 if sort else True, "target_limit 값은 10,000,000 이하여야 합니다."
        self.sort = sort
        self.target_limit = target_limit

    def value(self) -> Dict[str, Any]:
        return {"sort": self.sort, "target_limit": self.target_limit} if self.sort else {"sort": self.sort}


class CustomFeatureSourceConfig:
    def __init__(self, source_type: str, source_path: str):
        """
        예측 시 추가로 사용할 사용자 데이터 설정입니다(사용자 피쳐를 추가로 적용한 모델의 경우 필수).

        ## Args

        - source_type: (str) 데이터 타입 (`csv (deprecated)` | `edd2 (deprecated)` | `ye(deprecated)` | `file` | `table`).
        - source_path: (str) 데이터의 이름 또는 테이블 경로.
          - file: 파일 이름 (ex. test.csv)
          - table: BigQuery 테이블의 경로 (ex. sktaic-datahub.x1234567.test)

        ## Example

        ```python
        config = LabelDataConf(source_type="table", source_path="sktaic-datahub.x1234567.test")
        ```
        """
        assert source_type in [
            "csv",
            "edd2",
            "ye",
            "file",
            "table",
        ], "source_type 은(는) ['csv', 'edd2', 'ye', 'file', 'table'] 중 하나여야 합니다."

        if source_type in ["csv"]:
            print(
                "[Warn] CustomFeatureSourceConfig: 'csv' 타입은 deprecate 되었습니다. 'file' 타입을 사용해주세요.",
                file=sys.stderr,
            )
            source_type = "file"

        if source_type in ["edd2", "ye"]:
            print(
                "[Warn] CustomFeatureSourceConfig: 'edd2', 'ye' 타입은 deprecate 되었습니다. 'table' 타입을 사용해주세요.",
                file=sys.stderr,
            )
            source_type = "table"

        assert type(source_path) == str, "source_path 은(는) str 타입이어야 합니다."
        self.source_type = source_type
        self.source_path = source_path


class ExportChannelConfig:
    def __init__(self, channel: str):
        self.channel = channel


class EmailChannelConfig(ExportChannelConfig):
    param_name = "email_channel_config"

    def __init__(self, email: str):
        """
        이메일로 예측 결과를 전송하기 위한 설정 값입니다.

        ## Args

        - email: (str) 전송받을 이메일 경로 (@sktelecom.com 또는 @sk.com 도메인만 지원)

        ## Returns
        `sktmls.models.automl.EmailChannelConfig`

        ## Example

        ```python
        config = EmailChannelConfig("test@sktelecom.com")
        ```
        """
        super().__init__("EMAIL")
        assert type(email) == str, "email 은(는) str 타입이어야 합니다."
        self.email = email


class TOSChannelConfig(ExportChannelConfig):
    param_name = "tos_channel_config"

    def __init__(self, employee_num: str, target_group_name: str):
        """
        TOS로 예측 결과를 전송하기 위한 설정 값입니다.

        ## Args

        - employee_num: (str) 사번 (TOS 계정)
        - target_group_name: (str) 원하는 대상군 명

        ## Returns
        `sktmls.models.automl.TOSChannelConfig`

        ## Example

        ```python
        config = TOSChannelConfig("test@sktelecom.com")
        ```
        """
        super().__init__("TOS")
        assert type(employee_num) == str, "employee_num 은(는) str 타입이어야 합니다."
        assert type(target_group_name) == str, "target_group_name 은(는) str 타입이어야 합니다."
        self.employee_num = employee_num
        self.target_group_name = target_group_name


class EDDChannelConfig(ExportChannelConfig):
    param_name = "edd_channel_config"

    def __init__(self, username: str, database: str, table: str):
        """
        EDD로 예측 결과를 전송하기 위한 설정 값입니다.

        ## Args

        - username: (str) EDD 계정 (ex. g1234567)
        - database: (str) 데이터베이스 이름
        - table: (str) 테이블 이름

        ## Returns
        `sktmls.models.automl.EDDChannelConfig`

        ## Example

        ```python
        config = EDDChannelConfig("g1234567", "test_db", "test_table")
        ```
        """
        super().__init__("EDD")
        assert type(username) == str, "username 은(는) str 타입이어야 합니다."
        assert type(database) == str, "database 은(는) str 타입이어야 합니다."
        assert type(table) == str, "table 은(는) str 타입이어야 합니다."
        assert re.match(r"^[a-z][0-9]{7}", username), "잘못된 형식의 EDD 계정입니다."
        assert re.match(r"^[A-Za-z0-9_]+$", database), "잘못된 형식의 데이터베이스 이름입니다."
        assert re.match(r"^[A-Za-z0-9_]+$", table), "잘못된 형식의 테이블 이름입니다."
        self.username = username
        self.database = database
        self.table = table


class YEChannelConfig(ExportChannelConfig):
    param_name = "ye_channel_config"

    def __init__(self, username: str, table: str):
        """
        YE로 예측 결과를 전송하기 위한 설정 값입니다.

        ## Args

        - username: (str) YE 계정 (ex. "x1234567") 또는 BigQuery 데이터셋 이름 ("test_dataset")
        - table: (str) 테이블 이름 (ex. "test_table")

        ## Returns
        `sktmls.models.automl.YEChannelConfig`

        ## Example

        ```python
        config = YEChannelConfig("x1234567", "test_table")
        ```
        """
        super().__init__("YE")
        assert type(username) == str, "username 은(는) str 타입이어야 합니다."
        assert type(table) == str, "table 은(는) str 타입이어야 합니다."
        assert re.match(r"^[A-Za-z0-9_]+$", username), "잘못된 형식의 YE 계정 (또는 BigQuery 데이터셋 이름) 입니다."
        assert re.match(r"^[A-Za-z0-9_]+$", table), "잘못된 형식의 테이블 이름입니다."
        self.username = username
        self.table = table


class AutoMLBatchPrediction:
    def __init__(self, **kwargs):
        """
        AutoML 배치 예측 작업 클래스입니다.

        ## Args

        - id: (int) 예측 작업 ID.
        - channel: (string) 전송 채널 (`EMAIL` | `TOS` | `EDD` | `YE`)
        - status: (string) 예측 진행 상태 (`RUNNING` | `DONE` | `ERROR_BATCH_PREDICT`)
        - created_at: (datetime) 예측 시작 시간
        - updated_at: (datetime) 예측 완료 시간

        ## Returns
        `sktmls.models.automl.AutoMLBatchPrediction`
        """
        self.id = kwargs.get("id")
        self.channel = kwargs.get("name")
        self.status = kwargs.get("status")

        try:
            self.created_at = parser.parse(kwargs.get("created_at"))
        except TypeError:
            self.created_at = None

        try:
            self.updated_at = parser.parse(kwargs.get("updated_at"))
        except TypeError:
            self.updated_at = None

    def __str__(self) -> str:
        return str(self.id)


class AutoMLBatchPredictionClient(MLSClient):
    def __init__(
        self, env: MLSENV = None, runtime_env: MLSRuntimeENV = None, username: str = None, password: str = None
    ):
        """
        AutoML 배치 예측 관련 기능을 제공하는 클라이언트입니다.

        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - username: (str) MLS 계정명 (기본값: $MLS_USERNAME)
        - password: (str) MLS 계정 비밀번호 (기본값: $MLS_PASSWORD)

        아래의 환경 변수가 정의된 경우 해당 파라미터를 생략 가능합니다.

        - $MLS_ENV: env
        - $MLS_RUNTIME_ENV: runtime_env
        - $MLS_USERNAME: username
        - $MLS_PASSWORD: password

        ## Returns
        `sktmls.models.automl.AutoMLBatchPredictionClient`

        ## Example

        ```python
        client = AutoMLBatchPredictionClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")
        ```
        """
        super().__init__(env=env, runtime_env=runtime_env, username=username, password=password)

    def predict(
        self,
        model: AutoMLModel,
        export_channel_config: Union[EmailChannelConfig, TOSChannelConfig, EDDChannelConfig, YEChannelConfig],
        score_export_option: ScoreExportOption = None,
        period: str = None,
        custom_feature_source_config: CustomFeatureSourceConfig = None,
    ) -> AutoMLBatchPrediction:
        """
        배치 예측을 수행합니다.

        ## Args

        - model: (`sktmls.models.AutoMLModel`) 학습 모델
        - export_channel_config: (`sktmls.models.automl.ExportChannelConfig`) 예측 결과 전송 채널 설정
        - score_export_option: (`sktmls.models.automl.ScoreExportOption`) 스코어 전송 옵션 (스코어 기반 고객 타겟팅 문제의 경우 필수)
        - period: (str) 예측에 적용할 피쳐 스토어 기준연월 (형식: YYYYMM, 만약 입력하지 않을 경우 가장 최신의 기준연월을 적용)
        - custom_feature_source_config: (`sktmls.models.automl.CustomFeatureSourceConfig`) 예측에 적용할 커스텀 피쳐 데이터 파일 설정 (커스텀 피쳐를 사용한 모델의 경우 필수)

        ## Returns
        `sktmls.models.automl.AutoMLBatchPrediction`

        ## Example

        ```python
        client = AutoMLBatchPredictionClient(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE, username="mls_account", password="mls_password")

        model = model_client.get_automl_model(id=72)
        export_channel_config = YEChannelConfig(username="x1234567", table="triple_a_test")
        score_export_option = ScoreExportOption(sort=True, target_limit=100)
        custom_feature_config = CustomFeatureSourceConfig("ye", "my_custom_feature_table")

        prediction = client.predict(
            model=model,
            export_channel_config=export_channel_config,
            score_export_option=score_export_option,
            custom_feature_source_config=custom_feature_config,
        )
        ```
        """
        assert type(model) == AutoMLModel, "model 은(는) AutoMLModel 타입이어야 합니다."
        assert type(export_channel_config) in [
            EmailChannelConfig,
            TOSChannelConfig,
            EDDChannelConfig,
            YEChannelConfig,
        ], "export_channel_config 은(는) [EmailChannelConfig, TOSChannelConfig, EDDChannelConfig, YEChannelConfig] 중 하나여야 합니다."
        assert (
            score_export_option is None or type(score_export_option) == ScoreExportOption
        ), "score_export_option 은(는) ScoreExportOption 타입이어야 합니다."
        assert period is None or type(period) == str, "period 은(는) str 타입이어야 합니다."

        dataset_client = DatasetClient(
            self._MLSClient__env, self._MLSClient__runtime_env, self._MLSClient__username, self._MLSClient__password
        )
        dataset = dataset_client.get_dataset(id=model.dataset_id)

        if dataset.problem_type == ProblemType.SCORE and not score_export_option:
            raise AutoMLPredictionError("스코어 기반 고객 타겟팅 문제의 경우 score_export_option 파라미터가 필수입니다.")

        export_config = export_channel_config.__dict__
        channel = export_channel_config.channel
        del export_config["channel"]

        data = {"channel": channel}
        data[export_channel_config.param_name] = export_config

        if score_export_option:
            data["score_export_option"] = score_export_option.value()

        if period:
            data["period"] = period

        if custom_feature_source_config:
            data["custom_feature_source_config"] = custom_feature_source_config.__dict__

        return AutoMLBatchPrediction(
            **self._request(method="POST", url=f"/api/v1/models/{model.id}/predictions", data=data).results
        )



import setuptools


def load_long_description():
    with open("README.md", "r") as f:
        long_description = f.read()
    return long_description


def get_version():
    with open("sktmls/__init__.py", "r") as f:
        for line in f.readlines():
            if line.startswith("__version__"):
                return line.split('"')[1]
        else:
            raise TypeError("NO SKTMLS_VERSION")


setuptools.setup(
    name="sktmls",
    version=get_version(),
    author="SKTMLS",
    author_email="mls@sktai.io",
    description="MLS SDK",
    long_description=load_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/sktaiflow/mls-sdk",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
    install_requires=[
        "aiohttp[speedups]",
        "autogluon.tabular==1.4.0",
        "boto3",
        "catboost==1.2.8",
        "haversine==2.9.0",
        "joblib",
        "lightgbm==4.6.0",
        "numpy==2.3.3",
        "pandas==2.3.2",
        "pytz",
        "requests",
        "scikit-learn==1.7.2",
        "simplejson",
        "torch==2.8.0",
        "xgboost==3.0.5",
        "ipyparallel==9.0.1",
        "hvac==2.3.0",
        "distributed==2025.9.0",
        "dask==2025.9.0",
    ],
)



from pandas import DataFrame
from google.cloud.bigquery import PartitionRange, RangePartitioning, TimePartitioning
from google.cloud.bigquery.job import QueryJobConfig
from sktmls.models import MLSModelError
from typing import Dict
import hvac

CREDENTIALS_SECRET_PATH = "gcp/skt-datahub/dataflow"
PROJECT_ID = "skt-datahub"
TEMP_DATASET = "temp_1d"


def get_secrets(path, parse_data=True):
    vault_client = hvac.Client()
    data = vault_client.secrets.kv.v2.read_secret_version(path=path)
    if parse_data:
        data = data["data"]["data"]
    return data


def get_spark(scale=0, queue=None, jars=None):
    import os
    import uuid
    import tempfile
    from pyspark.sql import SparkSession
    from pyspark import version as spark_version

    is_spark_3 = spark_version.__version__ >= "3.0.0"

    tmp_uuid = str(uuid.uuid4())
    app_name = f"skt-{os.environ.get('USER', 'default')}-{tmp_uuid}"

    key = get_secrets("gcp/sktaic-datahub/dataflow")["config"]
    key_file_name = tempfile.mkstemp()[1]
    with open(key_file_name, "wb") as key_file:
        key_file.write(key.encode())
        os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = key_file.name

    if not queue:
        if "JUPYTERHUB_USER" in os.environ:
            queue = "dmig_eda"
        else:
            queue = "airflow_job"

    bigquery_jars = (
        "hdfs:///jars/spark-bigquery-with-dependencies_2.12-0.24.2.jar"
        if is_spark_3
        else "hdfs:///jars/spark-bigquery-with-dependencies_2.11-0.17.3.jar"
    )

    spark_jars = ",".join([bigquery_jars, jars]) if jars else bigquery_jars

    arrow_enabled = "spark.sql.execution.arrow.pyspark.enabled" if is_spark_3 else "spark.sql.execution.arrow.enabled"

    arrow_pre_ipc_format = "0" if is_spark_3 else "1"
    os.environ["ARROW_PRE_0_15_IPC_FORMAT"] = arrow_pre_ipc_format

    if queue == "nrt":
        spark = (
            SparkSession.builder.config("spark.app.name", app_name)
            .config("spark.driver.memory", "6g")
            .config("spark.executor.memory", "4g")
            .config("spark.driver.maxResultSize", "6g")
            .config("spark.rpc.message.maxSize", "1024")
            .config("spark.executor.core", "4")
            .config("spark.executor.instances", "32")
            .config("spark.yarn.queue", queue)
            .config("spark.ui.enabled", "false")
            .config("spark.port.maxRetries", "128")
            .config("spark.executorEnv.ARROW_PRE_0_15_IPC_FORMAT", arrow_pre_ipc_format)
            .config("spark.yarn.appMasterEnv.ARROW_PRE_0_15_IPC_FORMAT", arrow_pre_ipc_format)
            .config(
                "spark.jars",
                spark_jars,
            )
            .enableHiveSupport()
            .getOrCreate()
        )
        spark.conf.set(arrow_enabled, "true")
        return spark

    if scale in [1, 2, 3, 4]:
        spark = (
            SparkSession.builder.config("spark.app.name", app_name)
            .config("spark.driver.memory", f"{scale*8}g")
            .config("spark.executor.memory", f"{scale*3}g")
            .config("spark.executor.instances", f"{scale*8}")
            .config("spark.driver.maxResultSize", f"{scale*4}g")
            .config("spark.rpc.message.maxSize", "1024")
            .config("spark.yarn.queue", queue)
            .config("spark.ui.enabled", "false")
            .config("spark.port.maxRetries", "128")
            .config("spark.executorEnv.ARROW_PRE_0_15_IPC_FORMAT", arrow_pre_ipc_format)
            .config("spark.yarn.appMasterEnv.ARROW_PRE_0_15_IPC_FORMAT", arrow_pre_ipc_format)
            .config(
                "spark.jars",
                spark_jars,
            )
            .enableHiveSupport()
            .getOrCreate()
        )
    elif scale in [5, 6, 7, 8]:
        spark = (
            SparkSession.builder.config("spark.app.name", app_name)
            .config("spark.driver.memory", "8g")
            .config("spark.executor.memory", f"{2 ** scale}g")
            .config("spark.executor.instances", "32")
            .config("spark.driver.maxResultSize", "8g")
            .config("spark.rpc.message.maxSize", "1024")
            .config("spark.yarn.queue", queue)
            .config("spark.ui.enabled", "false")
            .config("spark.port.maxRetries", "128")
            .config("spark.executorEnv.ARROW_PRE_0_15_IPC_FORMAT", arrow_pre_ipc_format)
            .config("spark.yarn.appMasterEnv.ARROW_PRE_0_15_IPC_FORMAT", arrow_pre_ipc_format)
            .config(
                "spark.jars",
                spark_jars,
            )
            .enableHiveSupport()
            .getOrCreate()
        )
    else:
        if is_spark_3:
            spark = (
                SparkSession.builder.config("spark.app.name", app_name)
                .config("spark.driver.memory", "8g")
                .config("spark.executor.memory", "8g")
                .config("spark.executor.instances", "8")
                .config("spark.driver.maxResultSize", "6g")
                .config("spark.rpc.message.maxSize", "1024")
                .config("spark.yarn.queue", queue)
                .config("spark.ui.enabled", "false")
                .config("spark.port.maxRetries", "128")
                .config("spark.executorEnv.ARROW_PRE_0_15_IPC_FORMAT", arrow_pre_ipc_format)
                .config("spark.yarn.appMasterEnv.ARROW_PRE_0_15_IPC_FORMAT", arrow_pre_ipc_format)
                .config(
                    "spark.jars",
                    spark_jars,
                )
                .enableHiveSupport()
                .getOrCreate()
            )
        else:
            spark = (
                SparkSession.builder.config("spark.app.name", app_name)
                .config("spark.driver.memory", "6g")
                .config("spark.executor.memory", "8g")
                .config("spark.shuffle.service.enabled", "true")
                .config("spark.dynamicAllocation.enabled", "true")
                .config("spark.dynamicAllocation.maxExecutors", "200")
                .config("spark.driver.maxResultSize", "6g")
                .config("spark.rpc.message.maxSize", "1024")
                .config("spark.yarn.queue", queue)
                .config("spark.ui.enabled", "false")
                .config("spark.port.maxRetries", "128")
                .config("spark.executorEnv.ARROW_PRE_0_15_IPC_FORMAT", arrow_pre_ipc_format)
                .config("spark.yarn.appMasterEnv.ARROW_PRE_0_15_IPC_FORMAT", arrow_pre_ipc_format)
                .config(
                    "spark.jars",
                    spark_jars,
                )
                .enableHiveSupport()
                .getOrCreate()
            )
    spark.conf.set(arrow_enabled, "true")
    return spark


def _df_to_bq_table(
    df,
    dataset,
    table_name,
    partition=None,
    partition_field=None,
    clustering_fields=None,
    mode="overwrite",
    project_id=PROJECT_ID,
):
    import base64

    from skt.vault_utils import get_secrets

    key = get_secrets(CREDENTIALS_SECRET_PATH)["config"]
    table = f"{dataset}.{table_name}${partition}" if partition else f"{dataset}.{table_name}"
    df = (
        df.write.format("bigquery")
        .option("project", project_id)
        .option("credentials", base64.b64encode(key.encode()).decode())
        .option("table", table)
        .option("temporaryGcsBucket", "temp-seoul-7d")
    )
    if partition_field:
        df = df.option("partitionField", partition_field)
    if clustering_fields:
        df = df.option("clusteredFields", ",".join(clustering_fields))
    df.save(mode=mode)


def pandas_to_bq_table(
    pd_df,
    dataset,
    table_name,
    partition=None,
    partition_field=None,
    clustering_fields=None,
    mode="overwrite",
    project_id=PROJECT_ID,
):
    try:
        spark = get_spark()
        spark_df = spark.createDataFrame(pd_df)
        _df_to_bq_table(
            spark_df, dataset, table_name, partition, partition_field, clustering_fields, mode, project_id=project_id
        )
    finally:
        spark.stop()


def get_credentials():
    import json

    from google.oauth2 import service_account

    from skt.vault_utils import get_secrets

    key = get_secrets(CREDENTIALS_SECRET_PATH)["config"]
    json_acct_info = json.loads(key)
    credentials = service_account.Credentials.from_service_account_info(json_acct_info)
    scoped_credentials = credentials.with_scopes(["https://www.googleapis.com/auth/cloud-platform"])

    return scoped_credentials


def get_bigquery_client(credentials=None, project_id=PROJECT_ID):
    from google.cloud import bigquery

    if credentials is None:
        credentials = get_credentials()

    return bigquery.Client(credentials=credentials, project=project_id)


def load_query_result_to_partitions(query, dest_table, project_id=PROJECT_ID):
    from google.cloud.bigquery.dataset import DatasetReference  # noqa: F401
    from google.cloud.bigquery.table import TableReference  # noqa: F401

    bq = get_bigquery_client(project_id=project_id)
    table = bq.get_table(dest_table)

    """
    Destination 이 파티션일 때는 임시테이블 만들지 않고 직접 저장
    """
    if "$" in dest_table:
        qjc = QueryJobConfig(
            destination=table,
            write_disposition="WRITE_TRUNCATE",
            create_disposition="CREATE_IF_NEEDED",
            time_partitioning=table.time_partitioning,
            range_partitioning=table.range_partitioning,
            clustering_fields=table.clustering_fields,
        )
        job = bq.query(query, job_config=qjc)
        job.result()
        _print_query_job_results(job)
        return dest_table


def _print_query_job_results(query_job):
    try:
        t = query_job.destination
        dest_str = f"{t.project}.{t.dataset_id}.{t.table_id}" if t else "no destination"  # noqa: F841
    except Exception as e:
        print("Warning: exception on print statistics")
        print(e)


def bq_insert_overwrite_table(sql, destination, project_id=PROJECT_ID):
    bq = get_bigquery_client(project_id=project_id)
    table = bq.get_table(destination)
    if table.time_partitioning or table.range_partitioning:
        load_query_result_to_partitions(sql, destination, project_id)
    else:
        config = QueryJobConfig(
            destination=destination,
            write_disposition="WRITE_TRUNCATE",
            create_disposition="CREATE_NEVER",
            clustering_fields=table.clustering_fields,
        )
        job = bq.query(sql, config)
        job.result()
        _print_query_job_results(job)
        bq.close()


def get_temp_table(project_id=PROJECT_ID):
    import uuid

    table_id = str(uuid.uuid4()).replace("-", "_")
    full_table_id = f"{project_id}.{TEMP_DATASET}.{table_id}"

    return full_table_id


def _get_result_schema(sql, bq_client=None, project_id=PROJECT_ID):
    from google.cloud.bigquery.job import QueryJobConfig

    if bq_client is None:
        bq_client = get_bigquery_client(project_id=project_id)
    job_config = QueryJobConfig(
        dry_run=True,
        use_query_cache=False,
    )
    query_job = bq_client.query(sql, job_config=job_config)
    schema = query_job._properties["statistics"]["query"]["schema"]
    return schema


def _bq_query_to_new_table(sql, destination=None, project_id=PROJECT_ID):
    return bq_ctas(sql, destination, project_id=project_id)


def _get_result_column_type(sql, column, bq_client=None, project_id=PROJECT_ID):
    schema = _get_result_schema(sql, bq_client=bq_client, project_id=project_id)
    fields = schema["fields"]
    r = [field["type"] for field in fields if field["name"] == column]
    if r:
        return r[0]
    else:
        raise ValueError(f"Cannot find column {column} in {sql}")


def bq_ctas(sql, destination=None, partition_by=None, clustering_fields=None, project_id=PROJECT_ID):
    """
    create new table and insert results
    """
    from google.cloud.bigquery.job import QueryJobConfig

    bq = get_bigquery_client(project_id=project_id)
    if partition_by:
        partition_type = _get_result_column_type(sql, partition_by, bq_client=bq, project_id=project_id)
        if partition_type == "DATE":
            qjc = QueryJobConfig(
                destination=destination,
                write_disposition="WRITE_EMPTY",
                create_disposition="CREATE_IF_NEEDED",
                time_partitioning=TimePartitioning(field=partition_by),
                clustering_fields=clustering_fields,
            )
        elif partition_type == "INTEGER":
            qjc = QueryJobConfig(
                destination=destination,
                write_disposition="WRITE_EMPTY",
                create_disposition="CREATE_IF_NEEDED",
                range_partitioning=RangePartitioning(
                    PartitionRange(start=200001, end=209912, interval=1), field=partition_by
                ),
                clustering_fields=clustering_fields,
            )
        else:
            raise Exception(f"Partition column[{partition_by}] is neither DATE or INTEGER type.")
    else:
        qjc = QueryJobConfig(
            destination=destination,
            write_disposition="WRITE_EMPTY",
            create_disposition="CREATE_IF_NEEDED",
            clustering_fields=clustering_fields,
        )

    job = bq.query(sql, qjc)
    job.result()
    _print_query_job_results(job)
    bq.close()

    return job.destination


def _bq_table_to_pandas(table, project_id=PROJECT_ID):
    credentials = get_credentials()
    bq = get_bigquery_client(credentials=credentials, project_id=project_id)
    bqstorage_client = get_bigquery_storage_client(credentials=credentials)
    row_iterator = bq.list_rows(table)
    df = row_iterator.to_dataframe(bqstorage_client=bqstorage_client, progress_bar_type="tqdm")
    bq.close()

    return df


def get_bigquery_storage_client(credentials=None):
    from google.cloud import bigquery_storage

    if credentials is None:
        credentials = get_credentials()

    return bigquery_storage.BigQueryReadClient(credentials=credentials)


def bq_to_pandas(sql, large=False, project_id=PROJECT_ID):
    destination = None
    if large:
        destination = get_temp_table(project_id=project_id)
    destination = _bq_query_to_new_table(sql, destination, project_id=project_id)
    return _bq_table_to_pandas(destination, project_id=project_id)


class EurekaData:
    def __init__(
        self,
        model_name: str,
        model_version: str,
        features: Dict[str, list] = {},
        mls_feature_store_path: str = "di_mno_offer.mls_featurestore",
        mls_feature_store_meta_path: str = "di_mno_offer.mls_featurestore_meta",
    ):
        assert isinstance(model_name, str), "`model_name`은 str 타입이어야 합니다."
        assert isinstance(model_version, str), "`model_version`은 str 타입이어야 합니다."
        assert isinstance(mls_feature_store_path, str), "`mls_feature_store_path`은 str 타입이어야 합니다."
        assert isinstance(mls_feature_store_meta_path, str), "`mls_feature_store_meta_path`은 str 타입이어야 합니다."

        self.model_name = model_name
        self.model_version = model_version
        self.mls_feature_store_path = mls_feature_store_path
        self.mls_feature_store_meta_path = mls_feature_store_meta_path

        assert isinstance(features, dict), "`features`은 dict 타입이어야 합니다."
        for i in features:
            assert isinstance(features[i], list), "변수들은 list 타입이어야 합니다."
        if features:
            if sum(features.values(), []):  # 카테고리만 인입
                features = sum(features.values(), [])
            else:  # 변수도 함께 들어옴
                query = f"SELECT DISTINCT feature_nm FROM {PROJECT_ID}.{self.mls_feature_store_meta_path} WHERE feature_category IN ({str(list(features.keys()))[1:-1]})"
                meta = bq_to_pandas(query)
                features = meta["feature_nm"].values.tolist()

        else:
            query = f"SELECT DISTINCT feature_nm FROM {PROJECT_ID}.{self.mls_feature_store_meta_path}"  # 아무것도 안들어옴
            meta = bq_to_pandas(query)
            features = meta["feature_nm"].values.tolist()
        self.features = features

    def TrainPathLoader(
        self,
        target_df: DataFrame,
    ):
        assert isinstance(target_df, DataFrame), "`target_data`은 Pandas DataFrame 타입이어야 합니다."

        bq = get_bigquery_client()

        columns = ["user_id" if "svc_mgmt_num" == i else i for i in target_df.columns]
        target_df.columns = columns
        for i in ["user_id", "period"]:
            assert i in target_df.columns, f"target_df 에 `{i}` 컬럼이 없습니다."

        if "label" not in target_df.columns:
            target_df["label"] = 1

        try:
            target_df = target_df.astype({"label": "int"})
            target_df = target_df.astype({"period": "int"})
        except Exception as e:
            raise MLSModelError(f"EurekaModel: target_df의 타입 변환에 실패했습니다. {e}")

        try:
            pandas_to_bq_table(
                pd_df=target_df, dataset=TEMP_DATASET, table_name=f"{self.model_name}_{self.model_version}"
            )
        except Exception as e:
            raise MLSModelError(f"EurekaModel: target_df의 업로드에 실패했습니다. {e}")

        feature_list_query = ", ".join(self.features)
        max_ym = target_df["period"].max()
        if target_df["label"].nunique() == 1:
            assert target_df["label"].unique()[0] == 1, "`target_data`의 정답을 1개만 줄거라면 1만 입력되어야 합니다."

            try:
                bq.query(
                    f"DROP TABLE IF EXISTS {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}_label1"
                ).result()
                query = f"""
                CREATE TABLE IF NOT EXISTS {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}_label1 AS
                SELECT label_1.*
                  FROM (SELECT target.user_id
                             , target.label
                             , {feature_list_query}
                          FROM (SELECT * FROM {PROJECT_ID}.{self.mls_feature_store_path}) AS features
                         INNER JOIN (SELECT * FROM {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}) AS target
                            ON features.ym = target.period
                           AND features.svc_mgmt_num = target.user_id) AS label_1
                """
                bq.query(query).result()
                cnts = bq_to_pandas(
                    f"SELECT COUNT(*) FROM {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}_label1"
                )

                bq.query(
                    f"DROP TABLE IF EXISTS {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}_final"
                ).result()
                query = f"""
                CREATE TABLE IF NOT EXISTS {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}_final AS
                SELECT label_1.*
                  FROM {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}_label1 AS label_1
                 UNION DISTINCT  
                SELECT label_2.*
                  FROM (SELECT shuffle.svc_mgmt_num AS user_id
                             , shuffle.label
                             , {feature_list_query}
                          FROM (SELECT features.svc_mgmt_num
                                     , 0 AS label
                                     , {feature_list_query}
                                     , ROW_NUMBER() OVER (ORDER BY RAND()) AS rnd
                                  FROM (SELECT *
                                          FROM {PROJECT_ID}.{self.mls_feature_store_path}
                                         WHERE ym ={max_ym}) AS features
                                  LEFT JOIN (SELECT * FROM {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}) AS target
                                    ON features.svc_mgmt_num = target.user_id
                                 WHERE target.user_id IS NULL) AS shuffle
                         WHERE rnd <= {cnts.values[0][0]}) AS label_2
                """  # noqa: W291
                bq.query(query).result()
            except Exception as e:
                raise MLSModelError(f"EurekaModel: 데이터 생성에 실패했습니다. {e}")
        else:
            try:
                bq.query(
                    f"DROP TABLE IF EXISTS {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}_final"
                ).result()
                query = f"""
                CREATE TABLE {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}_final AS
                SELECT target.user_id
                     , target.label
                     , {feature_list_query}
                  FROM (SELECT *
                          FROM {PROJECT_ID}.{self.mls_feature_store_path}) AS features
                 INNER JOIN (SELECT * FROM {PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}) AS target
                    ON features.ym = target.period
                   AND features.svc_mgmt_num = target.user_id
                """
                bq.query(query).result()
            except Exception as e:
                raise MLSModelError(f"EurekaModel: 데이터 생성에 실패했습니다. {e}")

        self.destination = f"{PROJECT_ID}.{TEMP_DATASET}.{self.model_name}_{self.model_version}_final"
        return f"SELECT * FROM {self.destination}"

    def TrainDataLoader(
        self,
        target_df: DataFrame,
    ):
        assert isinstance(target_df, DataFrame), "`target_data`은 Pandas DataFrame 타입이어야 합니다."
        destination = self.TrainPathLoader(target_df)
        return bq_to_pandas(destination)

    def BatchLoader(self, periods, limit=None, filter_tbl=None):
        assert isinstance(periods, list), "`periods`은 List 타입이어야 합니다."
        feature_list_query = ", ".join(self.features)
        if limit:
            assert isinstance(limit, int), "`limit`은 int 타입이어야 합니다."
            if filter_tbl:
                assert isinstance(filter_tbl, str), "`filter_tbl`은 string 타입이어야 합니다."
                query = f"""SELECT T.svc_mgmt_num, {feature_list_query}
                FROM {PROJECT_ID}.{self.mls_feature_store_path} AS T
                INNER JOIN (SELECT svc_mgmt_num FROM {PROJECT_ID}.{filter_tbl}) AS T2
                ON T.svc_mgmt_num = T2.svc_mgmt_num
                INNER JOIN (SELECT DISTINCT TBL.svc_mgmt_num
                              FROM (SELECT svc_mgmt_num
                                         , row_number() over(partition by svc_mgmt_num order by op_dtm desc) as idx
                                      FROM wind.dbm_customer_agr
                                     WHERE dt = (SELECT max(dt) FROM wind.dbm_customer_agr)
                                       AND info_agr_itm_cd = '203'
                                       AND info_agr_cl_cd = '01') AS TBL
                             WHERE TBL.idx = 1) AS T3
                ON T.svc_mgmt_num = T3.svc_mgmt_num
                WHERE ym IN ({', '.join(periods)})
                LIMIT {limit}"""
            else:
                query = f"""SELECT T.svc_mgmt_num, {feature_list_query}
                FROM {PROJECT_ID}.{self.mls_feature_store_path} AS T
                INNER JOIN (SELECT DISTINCT TBL.svc_mgmt_num
                              FROM (SELECT svc_mgmt_num
                                         , row_number() over(partition by svc_mgmt_num order by op_dtm desc) as idx
                                      FROM wind.dbm_customer_agr
                                     WHERE dt = (SELECT max(dt) FROM wind.dbm_customer_agr)
                                       AND info_agr_itm_cd = '203'
                                       AND info_agr_cl_cd = '01') AS TBL
                             WHERE TBL.idx = 1) AS T3
                ON T.svc_mgmt_num = T3.svc_mgmt_num
                WHERE ym IN ({', '.join(periods)})
                LIMIT {limit}"""
        else:
            if filter_tbl:
                assert isinstance(filter_tbl, str), "`filter_tbl`은 string 타입이어야 합니다."
                query = f"""SELECT T.svc_mgmt_num, {feature_list_query}
                FROM {PROJECT_ID}.{self.mls_feature_store_path} AS T
                INNER JOIN (SELECT svc_mgmt_num FROM {PROJECT_ID}.{filter_tbl}) AS T2
                ON T.svc_mgmt_num = T2.svc_mgmt_num
                INNER JOIN (SELECT DISTINCT TBL.svc_mgmt_num
                              FROM (SELECT svc_mgmt_num
                                         , row_number() over(partition by svc_mgmt_num order by op_dtm desc) as idx
                                      FROM wind.dbm_customer_agr
                                     WHERE dt = (SELECT max(dt) FROM wind.dbm_customer_agr)
                                       AND info_agr_itm_cd = '203'
                                       AND info_agr_cl_cd = '01') AS TBL
                             WHERE TBL.idx = 1) AS T3
                ON T.svc_mgmt_num = T3.svc_mgmt_num
                WHERE ym IN ({', '.join(periods)})"""
            else:
                query = f"""SELECT T.svc_mgmt_num, {feature_list_query}
                FROM {PROJECT_ID}.{self.mls_feature_store_path} AS T
                INNER JOIN (SELECT DISTINCT TBL.svc_mgmt_num
                              FROM (SELECT svc_mgmt_num
                                         , row_number() over(partition by svc_mgmt_num order by op_dtm desc) as idx
                                      FROM wind.dbm_customer_agr
                                     WHERE dt = (SELECT max(dt) FROM wind.dbm_customer_agr)
                                       AND info_agr_itm_cd = '203'
                                       AND info_agr_cl_cd = '01') AS TBL
                             WHERE TBL.idx = 1) AS T3
                ON T.svc_mgmt_num = T3.svc_mgmt_num
                WHERE ym IN ({', '.join(periods)})"""
        return bq_to_pandas(query)
